import math
from typing import List, Literal, Optional, Union

from bolt11 import MilliSatoshi
from pydantic import BaseModel, Field, validator

from .exceptions import LnurlResponseException
from .types import (
    ClearnetUrl,
    DebugUrl,
    InitializationVector,
    LightningInvoice,
    LightningNodeUri,
    LnurlPayMetadata,
    Max144Str,
    OnionUrl,
)


class LnurlPayRouteHop(BaseModel):
    node_id: str = Field(..., alias="nodeId")
    channel_update: str = Field(..., alias="channelUpdate")


class LnurlPaySuccessAction(BaseModel):
    pass


class AesAction(LnurlPaySuccessAction):
    tag: Literal["aes"] = "aes"
    description: Max144Str
    ciphertext: str  # TODO
    iv: InitializationVector


class MessageAction(LnurlPaySuccessAction):
    tag: Literal["message"] = "message"
    message: Max144Str


class UrlAction(LnurlPaySuccessAction):
    tag: Literal["url"] = "url"
    url: Union[ClearnetUrl, OnionUrl, DebugUrl]
    description: Max144Str


class LnurlResponseModel(BaseModel):

    class Config:
        allow_population_by_field_name = True

    def dict(self, **kwargs):
        kwargs.setdefault("by_alias", True)
        return super().dict(**kwargs)

    def json(self, **kwargs):
        kwargs.setdefault("by_alias", True)
        return super().json(**kwargs)

    @property
    def ok(self) -> bool:
        return True


class LnurlErrorResponse(LnurlResponseModel):
    status: Literal["ERROR"] = "ERROR"
    reason: str

    @property
    def error_msg(self) -> str:
        return self.reason

    @property
    def ok(self) -> bool:
        return False


class LnurlSuccessResponse(LnurlResponseModel):
    status: Literal["OK"] = "OK"


class LnurlAuthResponse(LnurlResponseModel):
    tag: Literal["login"] = "login"
    callback: Union[ClearnetUrl, OnionUrl, DebugUrl]
    k1: str


class LnurlChannelResponse(LnurlResponseModel):
    tag: Literal["channelRequest"] = "channelRequest"
    uri: LightningNodeUri
    callback: Union[ClearnetUrl, OnionUrl, DebugUrl]
    k1: str


class LnurlHostedChannelResponse(LnurlResponseModel):
    tag: Literal["hostedChannelRequest"] = "hostedChannelRequest"
    uri: LightningNodeUri
    k1: str
    alias: Optional[str]


class LnurlPayResponse(LnurlResponseModel):
    tag: Literal["payRequest"] = "payRequest"
    callback: Union[ClearnetUrl, OnionUrl, DebugUrl]
    min_sendable: MilliSatoshi = Field(..., alias="minSendable", gt=0)
    max_sendable: MilliSatoshi = Field(..., alias="maxSendable", gt=0)
    metadata: LnurlPayMetadata

    @validator("max_sendable")
    def max_less_than_min(cls, value, values, **kwargs):  # noqa
        if "min_sendable" in values and value < values["min_sendable"]:
            raise ValueError("`max_sendable` cannot be less than `min_sendable`.")
        return value

    @property
    def min_sats(self) -> int:
        return int(math.ceil(self.min_sendable / 1000))

    @property
    def max_sats(self) -> int:
        return int(math.floor(self.max_sendable / 1000))


class LnurlPayResponseComment(LnurlPayResponse):
    """
    Adds the optional comment_allowed field to the LnurlPayResponse
    ref LUD-12: Comments in payRequest.
    """

    comment_allowed: int = Field(
        1000,
        description="Length of comment which can be sent",
        alias="commentAllowed",
    )


class LnurlPayActionResponse(LnurlResponseModel):
    pr: LightningInvoice
    success_action: Optional[Union[MessageAction, UrlAction, AesAction]] = Field(None, alias="successAction")
    routes: List[List[LnurlPayRouteHop]] = []


class LnurlWithdrawResponse(LnurlResponseModel):
    tag: Literal["withdrawRequest"] = "withdrawRequest"
    callback: Union[ClearnetUrl, OnionUrl, DebugUrl]
    k1: str
    min_withdrawable: MilliSatoshi = Field(..., alias="minWithdrawable", gt=0)
    max_withdrawable: MilliSatoshi = Field(..., alias="maxWithdrawable", gt=0)
    default_description: str = Field("", alias="defaultDescription")

    @validator("max_withdrawable")
    def max_less_than_min(cls, value, values, **kwargs):  # noqa
        if "min_withdrawable" in values and value < values["min_withdrawable"]:
            raise ValueError("`max_withdrawable` cannot be less than `min_withdrawable`.")
        return value

    @property
    def min_sats(self) -> int:
        return int(math.ceil(self.min_withdrawable / 1000))

    @property
    def max_sats(self) -> int:
        return int(math.floor(self.max_withdrawable / 1000))


class LnurlResponse:
    @staticmethod
    def from_dict(d: dict) -> LnurlResponseModel:
        try:
            if "tag" in d:
                # some services return `status` here, but it is not in the spec
                d.pop("status", None)

                return {
                    "channelRequest": LnurlChannelResponse,
                    "hostedChannelRequest": LnurlHostedChannelResponse,
                    "payRequest": LnurlPayResponse,
                    "withdrawRequest": LnurlWithdrawResponse,
                }[d["tag"]](**d)

            if "successAction" in d:
                d.pop("status", None)
                return LnurlPayActionResponse(**d)

            # some services return `status` in lowercase, but spec says upper
            d["status"] = d["status"].upper()

            if "status" in d and d["status"] == "ERROR":
                return LnurlErrorResponse(**d)

            return LnurlSuccessResponse(**d)

        except Exception:
            raise LnurlResponseException



import hashlib
import hmac
from io import BytesIO
from typing import List, Optional, Set, Tuple

from bech32 import bech32_decode, bech32_encode, convertbits
from ecdsa import SECP256k1, SigningKey

from .exceptions import InvalidLnurl, InvalidUrl


def lnurlauth_signature(domain: str, secret: str, k1: str) -> tuple[str, str]:
    hashing_key = hashlib.sha256(secret.encode()).digest()
    linking_key = hmac.digest(hashing_key, domain.encode(), "sha256")
    auth_key = SigningKey.from_string(linking_key, curve=SECP256k1, hashfunc=hashlib.sha256)
    sig = auth_key.sign_digest_deterministic(bytes.fromhex(k1), sigencode=encode_strict_der)
    assert auth_key.verifying_key, "LNURLauth verifying_key does not exist"
    key = auth_key.verifying_key.to_string("compressed")
    return key.hex(), sig.hex()


def _bech32_decode(bech32: str, *, allowed_hrp: Optional[Set[str]] = None) -> Tuple[str, List[int]]:
    hrp, data = bech32_decode(bech32)

    if not hrp or not data or (allowed_hrp and hrp not in allowed_hrp):
        raise ValueError(f"Invalid data or Human Readable Prefix (HRP): {hrp}.")

    return hrp, data


def _lnurl_clean(lnurl: str) -> str:
    lnurl = lnurl.strip()
    return lnurl.replace("lightning:", "") if lnurl.startswith("lightning:") else lnurl


def lnurl_decode(lnurl: str) -> str:
    """
    Decode a LNURL and return a url string without performing any validation on it.
    Use `lnurl.decode()` for validation and to get `Url` object.
    """
    _, data = _bech32_decode(_lnurl_clean(lnurl), allowed_hrp={"lnurl"})

    try:
        bech32_data = convertbits(data, 5, 8, False)
        assert bech32_data
        url = bytes(bech32_data).decode("utf-8")
        return url
    except UnicodeDecodeError:  # pragma: nocover
        raise InvalidLnurl


def url_encode(url: str) -> str:
    """
    Encode a URL without validating it first and return a bech32 LNURL string.
    Use `lnurl.encode()` for validation and to get a `Lnurl` object.
    """
    try:
        bech32_data = convertbits(url.encode("utf-8"), 8, 5, True)
        assert bech32_data
        lnurl = bech32_encode("lnurl", bech32_data)
    except UnicodeEncodeError:  # pragma: nocover
        raise InvalidUrl

    return lnurl.upper()


def _int_to_bytes_suitable_der(x: int) -> bytes:
    """for strict DER we need to encode the integer with some quirks"""
    b = x.to_bytes((x.bit_length() + 7) // 8, "big")

    if len(b) == 0:
        # ensure there's at least one byte when the int is zero
        return bytes([0])

    if b[0] & 0x80 != 0:
        # ensure it doesn't start with a 0x80 and so it isn't
        # interpreted as a negative number
        return bytes([0]) + b

    return b


def encode_strict_der(r: int, s: int, order: int):
    # if s > order/2 verification will fail sometimes
    # so we must fix it here see:
    # https://github.com/indutny/elliptic/blob/e71b2d9359c5fe9437fbf46f1f05096de447de57/lib/elliptic/ec/index.js#L146-L147
    if s > order // 2:
        s = order - s

    # now we do the strict DER encoding copied from
    # https://github.com/KiriKiri/bip66 (without any checks)
    r_temp = _int_to_bytes_suitable_der(r)
    s_temp = _int_to_bytes_suitable_der(s)

    r_len = len(r_temp)
    s_len = len(s_temp)
    sign_len = 6 + r_len + s_len

    signature = BytesIO()
    signature.write(0x30.to_bytes(1, "big", signed=False))
    signature.write((sign_len - 2).to_bytes(1, "big", signed=False))
    signature.write(0x02.to_bytes(1, "big", signed=False))
    signature.write(r_len.to_bytes(1, "big", signed=False))
    signature.write(r_temp)
    signature.write(0x02.to_bytes(1, "big", signed=False))
    signature.write(s_len.to_bytes(1, "big", signed=False))
    signature.write(s_temp)

    return signature.getvalue()



"""lnurl CLI"""

import asyncio
import sys

import click

from .core import encode as encode_lnurl
from .core import execute as execute_lnurl
from .core import handle as handle_lnurl
from .types import Lnurl

# disable tracebacks on exceptions
sys.tracebacklimit = 0


@click.group()
def command_group():
    """
    Python CLI for LNURL
    decode and encode lnurls"""


@click.command()
@click.argument("url", type=str)
def encode(url):
    """
    encode a URL
    """
    encoded = encode_lnurl(url)
    click.echo(encoded.bech32)


@click.command()
@click.argument("lnurl", type=str)
def decode(lnurl):
    """
    decode a LNURL
    """
    decoded = Lnurl(lnurl)
    click.echo(decoded.url)


@click.command()
@click.argument("lnurl", type=str)
def handle(lnurl):
    """
    handle a LNURL
    """
    decoded = asyncio.run(handle_lnurl(lnurl))
    click.echo(decoded.json())


@click.command()
@click.argument("lnurl", type=str)
@click.argument("msat_or_login", type=str, required=False)
def execute(lnurl, msat_or_login):
    """
    execute a LNURL request
    """
    if not msat_or_login:
        raise ValueError("You must provide either an amount_msat or a login_id.")
    res = asyncio.run(execute_lnurl(lnurl, msat_or_login))
    click.echo(res.json())


def main():
    """main function"""
    command_group.add_command(encode)
    command_group.add_command(decode)
    command_group.add_command(handle)
    command_group.add_command(execute)
    command_group()


if __name__ == "__main__":
    main()



from .core import decode, encode, execute, execute_login, execute_pay_request, get, handle
from .models import (
    LnurlAuthResponse,
    LnurlChannelResponse,
    LnurlErrorResponse,
    LnurlHostedChannelResponse,
    LnurlPayActionResponse,
    LnurlPayResponse,
    LnurlPayResponseComment,
    LnurlResponse,
    LnurlSuccessResponse,
    LnurlWithdrawResponse,
)
from .types import Lnurl

__all__ = [
    "decode",
    "encode",
    "execute",
    "execute_login",
    "execute_pay_request",
    "get",
    "handle",
    "Lnurl",
    "LnurlAuthResponse",
    "LnurlChannelResponse",
    "LnurlErrorResponse",
    "LnurlHostedChannelResponse",
    "LnurlPayActionResponse",
    "LnurlPayResponse",
    "LnurlPayResponseComment",
    "LnurlResponse",
    "LnurlSuccessResponse",
    "LnurlWithdrawResponse",
]



import json
import os
import re
from hashlib import sha256
from typing import List, Optional, Tuple, Union
from urllib.parse import parse_qs

from pydantic import (
    ConstrainedStr,
    HttpUrl,
    Json,
    ValidationError,
    parse_obj_as,
    validator,
)
from pydantic.errors import UrlHostTldError, UrlSchemeError
from pydantic.networks import Parts
from pydantic.validators import str_validator

from .exceptions import InvalidLnurlPayMetadata
from .helpers import _bech32_decode, _lnurl_clean, lnurl_decode


def ctrl_characters_validator(value: str) -> str:
    """Checks for control characters (unicode blocks C0 and C1, plus DEL)."""
    if re.compile(r"[\u0000-\u001f\u007f-\u009f]").search(value):
        raise ValueError
    return value


def strict_rfc3986_validator(value: str) -> str:
    """Checks for RFC3986 compliance."""
    if re.compile(r"[^]a-zA-Z0-9._~:/?#[@!$&'()*+,;=-]").search(value):
        raise ValueError
    return value


class ReprMixin:
    def __repr__(self) -> str:
        attrs = [  # type: ignore
            outer_slot  # type: ignore
            for outer_slot in [slot for slot in self.__slots__ if not slot.startswith("_")]  # type: ignore
            if getattr(self, outer_slot)  # type: ignore
        ]  # type: ignore
        extra = ", " + ", ".join(f"{n}={getattr(self, n).__repr__()}" for n in attrs) if attrs else ""
        return f"{self.__class__.__name__}({super().__repr__()}{extra})"


class Bech32(ReprMixin, str):
    """Bech32 string."""

    __slots__ = ("hrp", "data")

    def __new__(cls, bech32: str, **_) -> "Bech32":
        return str.__new__(cls, bech32)

    def __init__(self, bech32: str, *, hrp: Optional[str] = None, data: Optional[List[int]] = None):
        str.__init__(bech32)
        self.hrp, self.data = (hrp, data) if hrp and data else self.__get_data__(bech32)

    @classmethod
    def __get_data__(cls, bech32: str) -> Tuple[str, List[int]]:
        return _bech32_decode(bech32)

    @classmethod
    def __get_validators__(cls):
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value: str) -> "Bech32":
        hrp, data = cls.__get_data__(value)
        return cls(value, hrp=hrp, data=data)


class Url(HttpUrl):
    """URL with extra validations over pydantic's `HttpUrl`."""

    max_length = 2047  # https://stackoverflow.com/questions/417142/

    @classmethod
    def __get_validators__(cls):
        yield ctrl_characters_validator
        if os.environ.get("LNURL_STRICT_RFC3986", "0") == "1":
            yield strict_rfc3986_validator
        yield cls.validate

    @property
    def base(self) -> str:
        hostport = f"{self.host}:{self.port}" if self.port else self.host
        return f"{self.scheme}://{hostport}{self.path}"

    @property
    def query_params(self) -> dict:
        return {k: v[0] for k, v in parse_qs(self.query).items()}


class DebugUrl(Url):
    """Unsecure web URL, to make developers life easier."""

    allowed_schemes = {"http"}

    @classmethod
    def validate_host(cls, parts: Parts) -> Tuple[str, Optional[str], str, bool]:
        host, tld, host_type, rebuild = super().validate_host(parts)
        if host not in ["127.0.0.1", "0.0.0.0"]:
            raise UrlSchemeError()
        return host, tld, host_type, rebuild


class ClearnetUrl(Url):
    """Secure web URL."""

    allowed_schemes = {"https"}


class OnionUrl(Url):
    """Tor anonymous onion service."""

    allowed_schemes = {"https", "http"}

    @classmethod
    def validate_host(cls, parts: Parts) -> Tuple[str, Optional[str], str, bool]:
        host, tld, host_type, rebuild = super().validate_host(parts)
        if tld != "onion":
            raise UrlHostTldError()
        return host, tld, host_type, rebuild


class LightningInvoice(Bech32):
    """Bech32 Lightning invoice."""


class LightningNodeUri(ReprMixin, str):
    """Remote node address of form `node_key@ip_address:port_number`."""

    __slots__ = ("key", "ip", "port")

    def __new__(cls, uri: str, **kwargs) -> "LightningNodeUri":
        return str.__new__(cls, uri)

    def __init__(self, uri: str, *, key: Optional[str] = None, ip: Optional[str] = None, port: Optional[str] = None):
        str.__init__(uri)
        self.key = key
        self.ip = ip
        self.port = port

    @classmethod
    def __get_validators__(cls):
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value: str) -> "LightningNodeUri":
        try:
            key, netloc = value.split("@")
            ip, port = netloc.split(":")
        except Exception:
            raise ValueError

        return cls(value, key=key, ip=ip, port=port)


class Lnurl(ReprMixin, str):
    __slots__ = ("bech32", "url")

    def __new__(cls, lightning: str, **_) -> "Lnurl":
        return str.__new__(cls, _lnurl_clean(lightning))

    def __init__(self, lightning: str, *, url: Optional[Union[OnionUrl, ClearnetUrl, DebugUrl]] = None):
        bech32 = _lnurl_clean(lightning)
        str.__init__(bech32)
        self.bech32 = Bech32(bech32)
        self.url = url if url else self.__get_url__(bech32)

    @classmethod
    def __get_url__(cls, bech32: str) -> Union[OnionUrl, ClearnetUrl, DebugUrl]:
        url: str = lnurl_decode(bech32)
        return parse_obj_as(Union[OnionUrl, ClearnetUrl, DebugUrl], url)  # type: ignore

    @classmethod
    def __get_validators__(cls):
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value: str) -> "Lnurl":
        return cls(value, url=cls.__get_url__(value))

    @property
    def is_login(self) -> bool:
        return "tag" in self.url.query_params and self.url.query_params["tag"] == "login"


class LnAddress(ReprMixin, str):
    """Lightning address of form `user@host`"""

    def __new__(cls, address: str, **_) -> "LnAddress":
        return str.__new__(cls, address)

    def __init__(self, address: str):
        str.__init__(address)
        self.address = address
        self.url = self.__get_url__(address)

    @validator("address")
    def is_valid_email_address(cls, email: str) -> bool:
        email_regex = r"[A-Za-z0-9\._%+-]+@[A-Za-z0-9\.-]+\.[A-Za-z]{2,63}"
        return re.fullmatch(email_regex, email) is not None

    @classmethod
    def __get_url__(cls, address: str) -> Union[OnionUrl, ClearnetUrl, DebugUrl]:
        name, domain = address.split("@")
        url = ("http://" if domain.endswith(".onion") else "https://") + domain + "/.well-known/lnurlp/" + name
        return parse_obj_as(Union[OnionUrl, ClearnetUrl, DebugUrl], url)  # type: ignore


class LnurlPayMetadata(ReprMixin, str):
    valid_metadata_mime_types = {"text/plain", "image/png;base64", "image/jpeg;base64"}

    __slots__ = ("_list",)

    def __new__(cls, json_str: str, **kwargs) -> "LnurlPayMetadata":
        return str.__new__(cls, json_str)

    def __init__(self, json_str: str, *, json_obj: Optional[List] = None):
        str.__init__(json_str)
        self._list = json_obj if json_obj else self.__validate_metadata__(json_str)

    @classmethod
    def __validate_metadata__(cls, json_str: str) -> List[Tuple[str, str]]:
        try:
            parse_obj_as(Json[List[Tuple[str, str]]], json_str)
            data = [(str(item[0]), str(item[1])) for item in json.loads(json_str)]
        except ValidationError:
            raise InvalidLnurlPayMetadata

        clean_data = [x for x in data if x[0] in cls.valid_metadata_mime_types]
        mime_types = [x[0] for x in clean_data]
        counts = {x: mime_types.count(x) for x in mime_types}

        if not clean_data or "text/plain" not in mime_types or counts["text/plain"] > 1:
            raise InvalidLnurlPayMetadata

        return clean_data

    @classmethod
    def __get_validators__(cls):
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value: str) -> "LnurlPayMetadata":
        return cls(value, json_obj=cls.__validate_metadata__(value))

    @property
    def h(self) -> str:
        return sha256(self.encode("utf-8")).hexdigest()

    @property
    def text(self) -> str:
        output = ""

        for metadata in self._list:
            if metadata[0] == "text/plain":
                output = metadata[1]
                break

        return output

    @property
    def images(self) -> List[Tuple[str, str]]:
        return [x for x in self._list if x[0].startswith("image/")]

    def list(self) -> List[Tuple[str, str]]:
        return self._list


class InitializationVector(ConstrainedStr):
    min_length = 24
    max_length = 24


class Max144Str(ConstrainedStr):
    max_length = 144



class LnurlException(Exception):
    """A LNURL error occurred."""


class LnurlResponseException(LnurlException):
    """An error ocurred processing LNURL response."""


class InvalidLnurl(LnurlException, ValueError):
    """The LNURL provided was somehow invalid."""


class InvalidUrl(LnurlException, ValueError):
    """The URL is not properly formed."""


class InvalidLnurlPayMetadata(LnurlResponseException, ValueError):
    """The response `metadata` is not properly formed."""



from typing import Any, Optional, Union

import httpx
from bolt11 import Bolt11Exception, MilliSatoshi
from bolt11 import decode as bolt11_decode
from pydantic import ValidationError

from .exceptions import InvalidLnurl, InvalidUrl, LnurlResponseException
from .helpers import lnurlauth_signature, url_encode
from .models import LnurlAuthResponse, LnurlPayResponse, LnurlResponse, LnurlResponseModel, LnurlWithdrawResponse
from .types import ClearnetUrl, DebugUrl, LnAddress, Lnurl, OnionUrl


def decode(bech32_lnurl: str) -> Union[OnionUrl, ClearnetUrl, DebugUrl]:
    try:
        return Lnurl(bech32_lnurl).url
    except (ValidationError, ValueError):
        raise InvalidLnurl


def encode(url: str) -> Lnurl:
    try:
        return Lnurl(url_encode(url))
    except (ValidationError, ValueError):
        raise InvalidUrl


async def get(url: str, *, response_class: Optional[Any] = None) -> LnurlResponseModel:
    async with httpx.AsyncClient() as client:
        try:
            res = await client.get(url)
            res.raise_for_status()
        except Exception as e:
            raise LnurlResponseException(str(e))

        if response_class:
            assert issubclass(response_class, LnurlResponseModel), "Use a valid `LnurlResponseModel` subclass."
            return response_class(**res.json())

        return LnurlResponse.from_dict(res.json())


async def handle(
    bech32_lnurl: str,
    response_class: Optional[LnurlResponseModel] = None,
) -> LnurlResponseModel:
    try:
        if "@" in bech32_lnurl:
            lnaddress = LnAddress(bech32_lnurl)
            return await get(lnaddress.url, response_class=response_class)
        lnurl = Lnurl(bech32_lnurl)
    except (ValidationError, ValueError):
        raise InvalidLnurl

    if lnurl.is_login:
        return LnurlAuthResponse(callback=lnurl.url, k1=lnurl.url.query_params["k1"])

    return await get(lnurl.url, response_class=response_class)


async def execute(bech32_or_address: str, value: str) -> LnurlResponseModel:
    try:
        res = handle(bech32_or_address)
    except Exception as exc:
        raise LnurlResponseException(str(exc))

    if isinstance(res, LnurlPayResponse) and res.tag == "payRequest":
        return await execute_pay_request(res, value)
    elif isinstance(res, LnurlAuthResponse) and res.tag == "login":
        return await execute_login(res, value)
    elif isinstance(res, LnurlWithdrawResponse) and res.tag == "withdrawRequest":
        return await execute_withdraw(res, value)

    raise LnurlResponseException(f"{res.tag} not implemented")  # type: ignore


async def execute_pay_request(res: LnurlPayResponse, msat: str) -> LnurlResponseModel:
    if not res.min_sendable <= MilliSatoshi(msat) <= res.max_sendable:
        raise LnurlResponseException(f"Amount {msat} not in range {res.min_sendable} - {res.max_sendable}")
    try:
        async with httpx.AsyncClient() as client:
            res2 = await client.get(
                url=res.callback,
                params={
                    "amount": msat,
                },
            )
            res2.raise_for_status()
            return LnurlResponse.from_dict(res2.json())
    except Exception as exc:
        raise LnurlResponseException(str(exc))


async def execute_login(res: LnurlAuthResponse, secret: str) -> LnurlResponseModel:
    try:
        assert res.callback.host, "LNURLauth host does not exist"
        key, sig = lnurlauth_signature(res.callback.host, secret, res.k1)
        async with httpx.AsyncClient() as client:
            res2 = await client.get(
                url=res.callback,
                params={
                    "key": key,
                    "sig": sig,
                },
            )
            res2.raise_for_status()
            return LnurlResponse.from_dict(res2.json())
    except Exception as e:
        raise LnurlResponseException(str(e))


async def execute_withdraw(res: LnurlWithdrawResponse, pr: str) -> LnurlResponseModel:
    try:
        invoice = bolt11_decode(pr)
    except Bolt11Exception as exc:
        raise LnurlResponseException(str(exc))
    # if invoice does not have amount use the min withdrawable amount
    amount = invoice.amount_msat or res.min_withdrawable
    if not res.min_withdrawable <= MilliSatoshi(amount) <= res.max_withdrawable:
        raise LnurlResponseException(f"Amount {amount} not in range {res.min_withdrawable} - {res.max_withdrawable}")
    try:
        async with httpx.AsyncClient() as client:
            res2 = await client.get(
                url=res.callback,
                params={
                    "k1": res.k1,
                    "pr": pr,
                },
            )
            res2.raise_for_status()
            return LnurlResponse.from_dict(res2.json())
    except Exception as exc:
        raise LnurlResponseException(str(exc))
